import pickle
import re
import pytest
import sys
from mpi4py import MPI
from pytest_regressions.data_regression import DataRegressionFixture
from pytest_regressions.num_regression import NumericRegressionFixture
from pytest_regressions.file_regression import FileRegressionFixture


from pytimings import mpi

USE_MPI = [False]
if mpi.HAVE_MPI:
    USE_MPI.append(True)


@pytest.fixture(params=USE_MPI)
def use_mpi(request, monkeypatch):
    use_mpi = request.param
    if (not use_mpi) and mpi.HAVE_MPI:
        monkeypatch.delattr('mpi4py.MPI')
        monkeypatch.setattr('pytimings.mpi.HAVE_MPI', False)


@pytest.fixture
def timings_object(request, use_mpi):
    from pytimings.timer import Timings

    return Timings()


@pytest.fixture
def pickled_timings_object(request, use_mpi, shared_datadir):
    filename = f'nested.mpi_{mpi.HAVE_MPI}.pickle'
    return pickle.load(open(shared_datadir / filename, 'rb'))


def _make_regression_fixture(base_type, fname):
    def check(self, data_dict, basename=None, fullpath=None):
        """

        :param data_dict: the data to check
        :param basename: unless explicitly set, defaults to test name with rank count appended
        :param fullpath: part of the origianl API, but cannot be used (mutually exclusive in super class
            with setting basename
        """
        assert fullpath is None
        if basename is None:
            super_basename = re.sub(r"[\W]", "_", self.request.node.name)
            ranks = MPI.COMM_WORLD.Get_size()
            assert ranks > 0
            basename = f'{super_basename}_r{ranks}'
        print(f'checking basename {basename}')
        super(self.__class__, self).check(data_dict, basename=basename)

    new_type = type(f'Mpi{base_type.__name__}', (base_type,), {'check': check})

    @pytest.fixture()
    def maker(datadir, original_datadir, request):
        return new_type(datadir, original_datadir, request)

    maker.__name__ = fname
    return new_type, maker


for _rtype, _fname in (
    (DataRegressionFixture, 'data_regression'),
    (NumericRegressionFixture, 'num_regression'),
    (FileRegressionFixture, 'file_regression'),
):
    _fname = f'mpi_{_fname}'
    _new_type, _fixture = _make_regression_fixture(_rtype, _fname)
    locals()[_fname] = _fixture


def is_windows_platform():
    return sys.platform == 'win32' or sys.platform == 'cygwin'



#!/usr/bin/env python3

import sys
import os
import shutil

# left_marker = '_n'
# left_marker = '.01_'
left_marker = sys.argv[1]

for dirname in sys.argv[2:]:
    assert os.path.isdir(dirname)
    left = dirname.find(left_marker)
    right = dirname.find('_', left + len(left_marker))
    ranks = dirname[left + len(left_marker) : right]
    new = dirname.replace(left_marker + ranks, left_marker + ranks.zfill(8))
    print('mv {} {}'.format(dirname, new))
    shutil.move(dirname, new)



import atexit
import os
from functools import partial
from pathlib import Path
from typing import Union


def ensure_directory_exists(dirname):
    """create dirname if it doesn't exist"""
    try:
        os.makedirs(dirname)
    except FileExistsError:
        pass


def output_at_exit(output_dir: Union[str, Path] = None, csv_base='timings', timings=None) -> None:
    from pytimings.timer import global_timings

    output_dir = output_dir or os.getcwd()
    output_dir = Path(output_dir)
    ensure_directory_exists(output_dir)
    timings = timings or global_timings
    output = partial(timings.output_per_rank, output_dir=output_dir, csv_base=csv_base)
    atexit.register(output)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import matplotlib as mpl


def getHueVector(amount):
    level = 0
    while (1 << level) < amount:
        level += 1

    out = []
    return getHueVectorRec(out, amount, level)


def getHueVectorRec(out, amount, level):
    if level <= 1:
        if len(out) < amount:
            out.append(0.0)
        if len(out) < amount:
            out.append(0.5)
        return out
    else:
        out = getHueVectorRec(out, amount, level - 1)
        lower = len(out)
        out = getHueVectorRec(out, amount, level - 1)
        upper = len(out)
        for i in range(lower, upper):
            out[i] += 1.0 / (1 << level)
        return out


def getColourPalette(size):
    result = []  # colors
    huevector = []  # doubles
    satvalbifurcatepos = 0
    satvalsplittings = []  # doubles
    if len(satvalsplittings) == 0:
        # // insert ranges to bifurcate
        satvalsplittings.append(1)
        satvalsplittings.append(0)
        satvalbifurcatepos = 0

    huevector = getHueVector(size)
    bisectionlimit = 20
    for i in range(len(result), size):
        hue = huevector[i]
        saturation = 1
        value = 1
        switccolors = i % 3  # ; // why only 3 and not all combinations? because it's easy, plus the bisection limit
        # cannot be divided integer by it

        if i % bisectionlimit == 0:
            satvalbifurcatepos = satvalbifurcatepos % (len(satvalsplittings) - 1)
            toinsert = satvalbifurcatepos + 1
            satvalsplittings.insert(
                toinsert,
                (satvalsplittings[satvalbifurcatepos] - satvalsplittings[satvalbifurcatepos + 1]) / 2
                + satvalsplittings[satvalbifurcatepos + 1],
            )
            satvalbifurcatepos += 2

        if switccolors == 1:
            saturation = satvalsplittings[satvalbifurcatepos - 1]
        elif switccolors == 2:
            value = satvalsplittings[satvalbifurcatepos - 1]

        hue += 0.17  # ; // use as starting point a zone where color band is narrow so that small variations means
        # high change in visual effect
        if hue > 1:
            hue -= 1
        import colorsys

        col = colorsys.hsv_to_rgb(hue, saturation, value)
        result.append(col)
    return result


def contrast_ratio(color_a, color_b):
    def _lum(c):
        index = float(c) * 255
        if index < 0.03928:
            return index / 12.92
        else:
            return ((index + 0.055) / 1.055) ** 2.4

    def _rel(rgb):
        return 0.2126 * _lum(rgb[0]) + 0.7152 * _lum(rgb[1]) + 0.0722 * _lum(rgb[2])

    return (_rel(color_a) + 0.05) / (_rel(color_b) + 0.05)


def getColourPaletteCheat(size, filter_colors=None, bg_color=(1, 1, 1)):
    filter_colors = filter_colors or []
    k = []
    org_size = size
    while len(k) < org_size:
        size += 1
        k = [p for p in set(getColourPalette(size)) if p not in filter_colors]
        k = [p for p in set(getColourPalette(size)) if p not in filter_colors and contrast_ratio(p, bg_color) < 0.6]
    for p in k:
        print('{} ratio: {}'.format(p, contrast_ratio(p, bg_color)))
    return k


def discrete_cmap(count, bg_color=(255, 255, 255)):
    cmap = mpl.colors.ListedColormap(getColourPaletteCheat(count, bg_color=bg_color), 'indexed')
    mpl.cm.register_cmap(cmap=cmap)
    return cmap


if __name__ == '__main__':
    # k = getColourPalette( 9 )
    # print k
    # k = set(k)
    # print k
    print(getColourPalette(4))
    print(getColourPaletteCheat(4))
    print(getColourPaletteCheat(4, [(0.0, 0, 0.0)]))



#!/usr/bin/env python3


import sys
import os
import pandas_common as pc

common_string = pc.common_substring(sys.argv[1:])
merged = 'merged_{}.csv'.format(common_string)
baseline_name = 'msfem.all'

header, current = pc.read_files(sys.argv[1:])
headerlist = header['profiler']
current = pc.sorted_f(current, True)
current = pc.speedup(headerlist, current, baseline_name)
# pprint(t_sections)
current.transpose().to_csv(merged)
pc.plot_fem(current, merged)



from io import StringIO
from pytest_regressions import file_regression
import pytest

from .fixtures import timings_object, use_mpi, pickled_timings_object, mpi_file_regression


def test_output_all_measures(pickled_timings_object, mpi_file_regression):
    timing = pickled_timings_object
    with StringIO() as out:
        timing.output_all_measures(out)
        out.seek(0)
        mpi_file_regression.check(out.read())


def test_output_simple(pickled_timings_object, file_regression):
    timing = pickled_timings_object
    with StringIO() as out:
        timing.output_all_measures(out)
        out.seek(0)
        file_regression.check(out.read())


def test_output_per_rank(pickled_timings_object, file_regression, tmp_path):
    fn = pickled_timings_object.output_per_rank(output_dir=tmp_path, csv_base='per_rank')

    if fn is not None:
        # we're rank 0 and have written the summary file
        file_regression.check(open(fn).read())



#!/usr/bin/env python3
import logging
import sys

import pandas_common as pc

common_string = pc.common_substring(sys.argv[1:])
merged = 'merged_{}.csv'.format(common_string)
baseline_name = 'msfem.all'

header, current = pc.read_files(sys.argv[1:])
headerlist = header['profiler']
current = pc.sorted_f(current, True)
speedup = pc.speedup(headerlist, current, baseline_name)
# pprint(t_sections)
pc.plot_msfem(speedup, merged, original=current)
try:
    pc.plot_error(
        speedup,
        merged,
        ['msfem_exactH1s', 'msfem_exact_L2'],
        'grids.macro_cells_per_dim',
        ['$H^1_s$', '$L^2$', 'walltime'],
        baseline_name,
        logy_base=10,
    )
except KeyError:
    logging.error('No error data')
speedup.transpose().to_csv(merged)
# current.transpose().to_excel(merged+'.xls')
# current.to_csv(merged)



"""Main module."""
import functools
import shutil
import sys
import time
from contextlib import contextmanager

from io import StringIO
from collections import namedtuple, defaultdict
from pathlib import Path
import psutil
from typing import Dict, Tuple, Optional

from pytimings.mpi import get_communication_wrapper, get_local_communicator, get_communicator
from pytimings.tools import ensure_directory_exists

try:
    PERF_COUNTER_FUNCTION = time.perf_counter_ns
    THREAD_TIME_FUNCTION = time.time_ns
    TO_SECONDS_FACTOR = 1e-9
except AttributeError:
    PERF_COUNTER_FUNCTION = time.perf_counter
    try:
        THREAD_TIME_FUNCTION = time.thread_time
    except AttributeError:
        # TODO Log warning
        THREAD_TIME_FUNCTION = time.process_time
    TO_SECONDS_FACTOR = 1

THREAD_TIME = "thread"
WALL_TIME = "wall"
SYS_TIME = "sys"
USER_TIME = "user"

TimingDelta = namedtuple("TimingDelta", [WALL_TIME, SYS_TIME, USER_TIME], defaults=[0, 0, 0])


class NoTimerError(Exception):
    def __init__(self, section, timings=None):
        self.section = section
        self.timings = timings or global_timings
        avail = 'Available sections: ' + ','.join(self.timings._known_timers_map.keys())
        super().__init__(f"trying to access timer for unknown section '{section}'\n{avail}")


class TimingData:
    def __init__(self, name):
        self.name = name
        self._end_resources = None
        self._end_times = None
        self._process = psutil.Process()
        self._start_times = self._get()

    def _get(self):
        ps_times = self._process.cpu_times()
        return {
            USER_TIME: ps_times.user,
            SYS_TIME: ps_times.system,
            WALL_TIME: PERF_COUNTER_FUNCTION(),
            THREAD_TIME: THREAD_TIME_FUNCTION(),
        }

    def stop(self):
        self._end_times = self._get()

    def delta(self):
        delta_times = self._end_times or self._get()

        wall = (delta_times[WALL_TIME] - self._start_times[WALL_TIME]) * TO_SECONDS_FACTOR
        # kernel resource usage already is in seconds
        return TimingDelta(
            wall,
            delta_times[SYS_TIME] - self._start_times[SYS_TIME],
            delta_times[USER_TIME] - self._start_times[USER_TIME],
        )


def _default_timer_dict_entry():
    return (False, None)


class Timings:
    def __init__(self):
        self._commited_deltas: Dict[str, TimingDelta] = {}
        self._known_timers_map: Dict[str, Tuple[bool, Optional[TimingData]]] = defaultdict(_default_timer_dict_entry)
        self._csv_sep = ","
        self.reset()

    def start(self, section_name: str) -> None:
        """set this to begin a named section"""
        if section_name in self._known_timers_map.keys():
            running, data = self._known_timers_map[section_name]
            if running:
                # TODO log info
                return
        self._known_timers_map[section_name] = (True, TimingData(section_name))

    def stop(self, section_name: str = None) -> int:
        """stop named section's counter or all of them if section_name is None"""
        if section_name is None:
            for section in self._known_timers_map.keys():
                self.stop(section)
            return
        if section_name not in self._known_timers_map.keys():
            raise NoTimerError(section_name, self)
        self._known_timers_map[section_name] = (
            False,
            self._known_timers_map[section_name][1],
        )  # mark as not running
        timing = self._known_timers_map[section_name][1]
        timing.stop()
        delta = timing.delta()
        if section_name not in self._commited_deltas.keys():
            self._commited_deltas[section_name] = delta
        else:
            previous_delta = self._commited_deltas[section_name]
            new_delta = TimingDelta(*tuple(map(sum, zip(delta, previous_delta))))
            self._commited_deltas[section_name] = new_delta
        return self._commited_deltas[section_name]

    def reset(self, section_name: str = None) -> None:
        """set elapsed time back to 0 for a given section or all of them if section_name is None"""
        if section_name is None:
            for section in self._known_timers_map.keys():
                self.reset(section)
            return
        try:
            self.stop(section_name)
        except NoTimerError:
            pass  # not stopping no timer is not an error
        self._commited_deltas[section_name] = TimingDelta()

    def walltime(self, section_name: str) -> int:
        """get runtime of section in milliseconds"""
        return self.delta(section_name)[WALL_TIME]

    def delta(self, section_name: str) -> Dict[str, int]:
        """get the full delta dict"""
        try:
            return self._commited_deltas[section_name]
        except KeyError:
            raise NoTimerError(section_name, self)

    def output_per_rank(self, output_dir: Path, csv_base: str) -> None:
        """
             creates one file local to each MPI-rank (no global averaging)
        *  one single rank-0 file with all combined/averaged measures
        """
        communication = get_communication_wrapper()
        rank = communication.rank

        ensure_directory_exists(output_dir)
        filename = output_dir / f"{csv_base}_p{rank:08d}.csv"
        with open(filename, "wt") as outfile:
            self.output_all_measures(outfile, get_local_communicator())
        tmp_out = StringIO()
        # all ranks have to participate in the data generation
        self.output_all_measures(tmp_out, get_communicator())
        # but only rank 0 needs to write it
        if rank == 0:
            a_filename = output_dir / f"{csv_base}.csv"
            tmp_out.seek(0)
            open(a_filename, "wt").write(tmp_out.read())
            return a_filename

    def output_simple(self, out=None):
        """outputs walltime only w/o MPI-rank averaging"""
        out = out or sys.stdout
        for section in self._commited_deltas.keys:
            out.write(f"{self._csv_sep}{section}")
        for delta in self._commited_deltas.values():
            out.write(f"{self._csv_sep}{delta[WALL_TIME]}")

    def output_all_measures(self, out=None, mpi_comm=None) -> None:
        """output all recorded measures
        * \note outputs average, min, max over all MPI processes associated to mpi_comm **/"""
        out = out or sys.stdout
        mpi_comm = mpi_comm or get_communicator()
        comm = get_communication_wrapper(mpi_comm)
        stash = StringIO()
        sep = self._csv_sep
        stash.write(f"threads{sep}ranks")
        for section in self._commited_deltas.keys():
            stash.write(
                f'{sep}{section}_avg_usr{sep}{section}_max_usr{sep}{section}_avg_wall{sep}{section}_max_wall{sep}{section}_avg_sys{sep}{section}_max_sys'
            )

        weight = 1 / comm.size

        # threadManager().max_threads() {sep} comm.size()
        stash.write(f"\n1{sep}{comm.size}")

        for section, delta in self._commited_deltas.items():
            delta = delta._asdict()
            wall = delta[WALL_TIME]
            usr = delta[USER_TIME]
            syst = delta[SYS_TIME]
            wall_sum = comm.sum(wall)
            wall_max = comm.max(wall)
            usr_sum = comm.sum(usr)
            usr_max = comm.max(usr)
            sys_sum = comm.sum(syst)
            sys_max = comm.max(syst)
            stash.write(
                f"{sep}{usr_sum * weight}{sep}{usr_max}{sep}{wall_sum * weight}{sep}{wall_max}{sep}{sys_sum * weight}{sep}{sys_max}"
            )

        stash.write("\n")
        stash.seek(0)
        if comm.rank == 0:
            shutil.copyfileobj(stash, out)


global_timings = Timings()


@contextmanager
def scoped_timing(section_name, log_function=None, timings=None):
    timings = timings or global_timings
    timings.start(section_name)
    try:
        yield
    finally:
        delta = timings.stop(section_name)
        if log_function:
            log_function(f"Executing {section_name} took {delta.wall * TO_SECONDS_FACTOR}s\n")


def function_timer(section_name, log_function=None, timings=None):
    def decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            with scoped_timing(section_name=section_name, log_function=log_function, timings=timings):
                return function(*args, **kwargs)

        return wrapper

    return decorator



#!/usr/bin/env python

"""Tests for `pytimings` package."""
import pickle
import sys
import time
from functools import partial
from tempfile import TemporaryFile

from click.testing import CliRunner

from pytimings import cli, mpi
from pytimings.timer import scoped_timing, function_timer
from .fixtures import timings_object, use_mpi, pickled_timings_object, is_windows_platform

_DUMMY_SECTION = 'mysection'

DEFAULT_SLEEP_SECONDS = 0.1


def _busywait(secs):
    """busywait simulates load, so user time won't be 0 in timings"""
    init_time = time.time()
    while time.time() < init_time + secs:
        pass
    return time.time() - init_time


default_sleep = partial(_busywait, DEFAULT_SLEEP_SECONDS)


def _assert(delta_value, lower=DEFAULT_SLEEP_SECONDS, upper=None):
    if is_windows_platform():
        fuzzy_factor = 0.99
        lower = lower * fuzzy_factor
        if upper:
            upper = upper / fuzzy_factor
    if upper is None:
        assert delta_value > lower
    else:
        assert upper > delta_value > lower


def test_content(timings_object):
    timings_object.start(_DUMMY_SECTION)
    slept = default_sleep()
    timed = timings_object.stop(_DUMMY_SECTION)
    _assert(timed.wall, lower=slept)
    timings_object.stop()
    timings_object.output_all_measures()
    timings_object.reset(_DUMMY_SECTION)
    timings_object.reset()
    timings_object.start(_DUMMY_SECTION)
    slept = default_sleep()
    timed = timings_object.stop(_DUMMY_SECTION)
    _assert(timed.wall, lower=slept, upper=1.1 * slept)
    timings_object.stop()
    timings_object.output_all_measures()


def test_nesting(timings_object):
    scope = partial(scoped_timing, timings=timings_object)
    sleeps = []
    with scope("root_section"):
        sleeps.append(default_sleep())
        with scope("root_section.nested_1"):
            sleeps.append(default_sleep())
            with scope("root_section.nested_1.leaf_section"):
                sleeps.append(default_sleep())
            timed = timings_object.delta("root_section.nested_1.leaf_section")
            _assert(timed.wall, lower=sum(sleeps[-1:]), upper=1.1 * sum(sleeps[-2:]))
        timed = timings_object.delta("root_section.nested_1")
        _assert(timed.wall, lower=sum(sleeps[-2:]), upper=1.1 * sum(sleeps[-3:]))
    timed = timings_object.delta("root_section")
    _assert(timed.wall, lower=sum(sleeps), upper=1.1 * sum(sleeps))


def test_pickling(pickled_timings_object):
    with TemporaryFile() as out:
        dump = partial(pickle.dump, pickled_timings_object, out)
        dump(protocol=0)
        dump(protocol=pickle.HIGHEST_PROTOCOL)
        dump(protocol=pickle.DEFAULT_PROTOCOL)


def test_context(timings_object):
    timings_object.start(_DUMMY_SECTION)
    default_sleep()
    timings_object.stop(_DUMMY_SECTION)
    delta_before = timings_object.delta(_DUMMY_SECTION)
    with scoped_timing(_DUMMY_SECTION, timings=timings_object):
        default_sleep()
    delta_after = timings_object.delta(_DUMMY_SECTION)
    _assert(delta_after.wall, lower=delta_before.wall)
    _assert(delta_after.user, lower=delta_before.user)

    # use the global timings object default
    delta_before = timings_object.delta(_DUMMY_SECTION)
    with scoped_timing(_DUMMY_SECTION):
        default_sleep()
    delta_after = timings_object.delta(_DUMMY_SECTION)
    assert delta_after == delta_before


def test_decorator(timings_object):
    @function_timer(section_name=_DUMMY_SECTION, timings=timings_object)
    def decorated():
        default_sleep()

    assert _DUMMY_SECTION not in timings_object._known_timers_map.keys()
    decorated()
    delta_before = timings_object.delta(_DUMMY_SECTION)
    decorated()
    delta_after = timings_object.delta(_DUMMY_SECTION)
    _assert(delta_after.wall, lower=delta_before.wall)
    _assert(delta_after.user, lower=delta_before.user)


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    assert 'pytimings.cli.main' in result.output
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output



try:
    from mpi4py import MPI

    HAVE_MPI = True
except ImportError:
    HAVE_MPI = False


class _SerialCommunicationWrapper:
    """Minimal no-MPI abstraction around needed collective communication operations

    This class should not be instantiated directly. Use `get_communication_wrapper instead`"""

    def __init__(self, mpi_communicator=None):
        if mpi_communicator is not None:
            # TODO log debug
            pass
        self._comm = mpi_communicator
        self.rank = 0
        self.size = 1

    def sum(self, local_value):
        return local_value

    def max(self, local_value):
        return local_value


class DummyCommunicator:
    rank = 0
    size = 1


if HAVE_MPI:

    class _MPICommunicationWrapper:
        """Minimal MPI abstraction around needed collective communication operations

        This class should not be instantiated directly. Use `get_communication_wrapper instead`"""

        def __init__(self, mpi_communicator=None):
            self._comm = mpi_communicator or get_communicator()
            self.rank = self._comm.rank
            self.size = self._comm.size

        def sum(self, local_value):
            lsum = self._comm.reduce(local_value, MPI.SUM)
            return lsum

        def max(self, local_value):
            lsum = self._comm.reduce(local_value, MPI.MAX)
            return lsum

    _CommunicationWrapper = _MPICommunicationWrapper
else:

    _CommunicationWrapper = _SerialCommunicationWrapper


def get_communication_wrapper(mpi_communicator=None):
    if HAVE_MPI:
        return _MPICommunicationWrapper(mpi_communicator)
    return _SerialCommunicationWrapper(mpi_communicator)


def get_local_communicator():
    if HAVE_MPI:
        return MPI.COMM_SELF
    return DummyCommunicator()


def get_communicator():
    if HAVE_MPI:
        return MPI.COMM_WORLD
    return DummyCommunicator()



import math

__author__ = 'r_milk01'

import os
import pandas as pd
from configparser import ConfigParser

import matplotlib.pyplot as plt
import matplotlib
import itertools
import logging
import difflib
import colors as color_util

TIMINGS = ['usr', 'sys', 'wall']
MEASURES = ['max', 'avg']
SPECIALS = ['run', 'threads', 'ranks', 'cores']
'''markers = {0: u'tickleft', 1: u'tickright', 2: u'tickup', 3: u'tickdown', 4: u'caretleft', u'D': u'diamond',
           6: u'caretup', 7: u'caretdown', u's': u'square', u'|': u'vline', u'': u'nothing', u'None': u'nothing',
           None: u'nothing', u'x': u'x', 5: u'caretright', u'_': u'hline', u'^': u'triangle_up', u' ': u'nothing',
           u'd': u'thin_diamond', u'h': u'hexagon1', u'+': u'plus', u'*': u'star', u',': u'pixel', u'o': u'circle',
           u'.': u'point', u'1': u'tri_down', u'p': u'pentagon', u'3': u'tri_left', u'2': u'tri_up', u'4': u'tri_right',
           u'H': u'hexagon2', u'v': u'triangle_down', u'8': u'octagon', u'<': u'triangle_left', u'>': u'triangle_right'}
'''
MARKERS = ['s', 'o', 4, 5, 7, '|', '*', 1, 2, 3, 4, 6, 7]
FIGURE_OUTPUTS = ['png', 'pdf', 'pgf']

# pd.options.display.mpl_style = 'default'
# matplotlib.rc('font', family='sans-serif')
# matplotlib.rc('xtick', labelsize=20)
# matplotlib.rc('ytick', labelsize=20)
SMALL_SIZE = 11
MEDIUM_SIZE = 13
BIGGER_SIZE = 16
matplotlib.rc('font', size=MEDIUM_SIZE, family='sans-serif')  # controls default text sizes
matplotlib.rc('axes', titlesize=BIGGER_SIZE)  # fontsize of the axes title
matplotlib.rc('axes', labelsize=BIGGER_SIZE)  # fontsize of the x and y labels
matplotlib.rc('xtick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
matplotlib.rc('ytick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
matplotlib.rc('legend', fontsize=SMALL_SIZE)  # legend fontsize
matplotlib.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

# http://nerdjusttyped.blogspot.de/2010/07/type-1-fonts-and-matplotlib-figures.html
matplotlib.rcParams['ps.useafm'] = True
matplotlib.rcParams['text.usetex'] = True
matplotlib.rcParams['pgf.texsystem'] = 'pdflatex'


def common_substring(strings, glue='_'):
    first, last = strings[0], strings[-1]
    seq = difflib.SequenceMatcher(None, first, last, autojunk=False)
    mb = seq.get_matching_blocks()
    return glue.join([first[m.a : m.a + m.size] for m in mb]).replace(os.path.sep, '')


def make_val(val, round_digits=3):
    try:
        return round(float(val), round_digits)
    except ValueError:
        return str(val)


def m_strip(s, timings=None, measures=None):
    timings = timings or TIMINGS
    measures = measures or MEASURES
    for t, m in itertools.product(timings, measures):
        s = s.replace('_{}_{}'.format(m, t), '')
    return s


def read_files(dirnames, specials=None):
    current = None
    specials = specials or SPECIALS
    header = {'memory': [], 'profiler': [], 'params': [], 'errors': []}
    for fn in dirnames:
        assert os.path.isdir(fn)
        prof = os.path.join(fn, 'profiler.csv')
        try:
            new = pd.read_csv(prof)
        except pd.parser.CParserError as e:
            logging.error('Failed parsing {}'.format(prof))
            raise e
        header['profiler'] = list(new.columns.values)
        params = ConfigParser()
        param_fn = ['dsc_parameter.log', 'dxtc_parameter.log']
        subdirs = ['', 'logs', 'logdata']
        params.read([os.path.join(fn, sd, pfn) for sd, pfn in itertools.product(subdirs, param_fn)])
        p = {}
        for section in params.sections():
            p.update({'{}.{}'.format(section, n): make_val(v) for n, v in params.items(section)})
        p['grids.total_macro_cells'] = math.pow(p['grids.macro_cells_per_dim'], p['grids.dim'])
        p['grids.total_fine_cells'] = p['grids.total_macro_cells'] * math.pow(
            p['grids.micro_cells_per_macrocell_dim'], p['grids.dim']
        )
        param = pd.DataFrame(p, index=[0])
        # mem
        mem = os.path.join(fn, 'memory.csv')
        mem = pd.read_csv(mem)
        new = pd.concat((new, param, mem), axis=1)
        header['memory'] = mem.columns.values
        header['params'] = param.columns.values
        err = os.path.join(fn, 'errors.csv')
        if os.path.isfile(err):
            err = pd.read_csv(err)
            header['errors'] = err.columns.values
            new = pd.concat((new, err), axis=1)

        current = current.append(new, ignore_index=True) if current is not None else new
    # ideal speedup account for non-uniform thread/rank ratio across columns
    count = len(current['ranks'])
    cmp_value = lambda j: current['grids.total_macro_cells'][j] / (current['ranks'][j] * current['threads'][j])
    values = [cmp_value(i) / cmp_value(0) for i in range(0, count)]
    current.insert(len(specials), 'ideal_scaleup', pd.Series(values))
    cmp_value = lambda j: current['ranks'][j] * current['threads'][j]
    values = [cmp_value(i) / cmp_value(0) for i in range(0, count)]
    current.insert(len(specials), 'ideal_speedup', pd.Series(values))
    cores = [cmp_value(i) for i in range(0, count)]
    current.insert(len(specials), 'cores', pd.Series(cores))
    cmp_value = lambda j: current['grids.total_macro_cells'][j] / (current['ranks'][j] * current['threads'][j])
    values = [cmp_value(i) / cmp_value(0) for i in range(0, count)]
    current.insert(len(specials), 'ideal_time', pd.Series(values))

    return header, current


def sorted_f(frame, ascending=True, sort_cols=None):
    sort_cols = sort_cols or ['ranks', 'threads']
    return frame.sort_values(by=sort_cols, na_position='last', ascending=ascending)


def speedup(headerlist, current, baseline_name, specials=None, round_digits=3, timings=None, measures=None):
    timings = timings or TIMINGS
    measures = measures or MEASURES
    specials = specials or SPECIALS
    t_sections = set([m_strip(h) for h in headerlist]) - set(specials)

    for sec in t_sections:
        for t, m in itertools.product(timings, measures):
            source_col = '{}_{}_{}'.format(sec, m, t)
            source = current[source_col]

            speedup_col = source_col + '_speedup'
            ref_value = source[0]
            values = [round(ref_value / source[i], round_digits) for i in range(len(source))]
            current[speedup_col] = pd.Series(values)

            # relative part of overall absolut timing category
            abspart_col = source_col + '_abspart'
            ref_value = lambda j: float(current['{}_{}_{}'.format(baseline_name, m, t)][j])
            values = [round(source[i] / ref_value(i), round_digits) for i in range(len(source))]
            current[abspart_col] = pd.Series(values)

            # relative part of overall total walltime
            wallpart_col = source_col + '_wallpart'
            ref_value = lambda j: float(current['{}_{}_{}'.format(baseline_name, m, 'wall')][j])
            values = [round(source[i] / ref_value(i), round_digits) for i in range(len(source))]
            current[wallpart_col] = pd.Series(values)

        for m in measures:
            # thread efficiency
            source_col = '{}_{}_{}'.format(sec, m, 'usr')
            threadeff_col = source_col + '_threadeff'
            wall = current['{}_{}_{}'.format(sec, m, 'wall')]
            source = current[source_col]
            value = lambda j: float(source[j] / (current['threads'][j] * wall[j]))
            values = [round(value(i), round_digits) for i in range(len(source))]
            current[threadeff_col] = pd.Series(values)

    current = sorted_f(current, True)
    return current


def scaleup(headerlist, current, baseline_name, specials=None, round_digits=3, timings=None, measures=None):
    timings = timings or TIMINGS
    measures = measures or MEASURES
    specials = specials or SPECIALS
    t_sections = set([m_strip(h) for h in headerlist]) - set(specials)

    for sec in t_sections:
        for t, m in itertools.product(timings, measures):
            source_col = '{}_{}_{}'.format(sec, m, t)
            source = current[source_col]

            speedup_col = '{}_{}'.format(source_col, 'scaleup')
            ref_value = source[0]
            values = [round(ref_value / source[i], round_digits) for i in range(len(source))]
            current[speedup_col] = pd.Series(values)

            # relative part of overall absolut timing category
            abspart_col = source_col + '_abspart'
            ref_value = lambda j: float(current['{}_{}_{}'.format(baseline_name, m, t)][j])
            values = [round(source[i] / ref_value(i), round_digits) for i in range(len(source))]
            current[abspart_col] = pd.Series(values)

            # relative part of overall total walltime
            wallpart_col = source_col + '_wallpart'
            ref_value = lambda j: float(current['{}_{}_{}'.format(baseline_name, m, 'wall')][j])
            values = [round(source[i] / ref_value(i), round_digits) for i in range(len(source))]
            current[wallpart_col] = pd.Series(values)

        for m in measures:
            # thread efficiency
            source_col = '{}_{}_{}'.format(sec, m, 'usr')
            threadeff_col = source_col + '_threadeff'
            wall = current['{}_{}_{}'.format(sec, m, 'wall')]
            source = current[source_col]
            value = lambda j: float(source[j] / (current['threads'][j] * wall[j]))
            values = [round(value(i), round_digits) for i in range(len(source))]
            current[threadeff_col] = pd.Series(values)

    current = sorted_f(current, True)
    return current


def plot_msfem(current, filename_base, series_name=None, xcol=None, original=None):
    xcol = xcol or 'cores'
    series_name = series_name or 'speedup'
    categories = ['Elliptic_MsFEM_Solver.apply', 'coarse.solve', 'local.solve_for_all_cells', 'coarse.assemble']
    measure = 'usr'
    ycols = ['msfem.{}_avg_{}_{}'.format(v, measure, series_name) for v in categories] + [
        'ideal_{}'.format(series_name)
    ]
    bar_cols = ['msfem.{}_avg_{}_abspart'.format(v, measure) for v in categories[1:]]
    labels = ['Overall', 'Coarse solve', 'Local assembly + solve', 'Coarse assembly'] + ['Ideal']
    plot_common(
        current,
        filename_base,
        ycols,
        labels,
        bar=(bar_cols, ['Coarse solve', 'Local assembly + solve', 'Coarse assembly']),
        xcol=xcol,
        series_name=series_name,
        logx_base=2,
        logy_base=2,
    )

    ycols = ['msfem.{}_avg_{}'.format(v, measure) for v in categories]
    for col in ycols:
        original[col] = original[col] / 1000.0
    bar_cols = ['msfem.{}_avg_{}_abspart'.format(v, measure) for v in categories[1:]]
    series_name = 'Time (s)'
    labels = ['Overall', 'Coarse solve', 'Local assembly + solve', 'Coarse assembly'] + ['Ideal']
    plot_common(
        original,
        filename_base,
        ycols,
        labels,
        bar=(bar_cols, ['Coarse solve', 'Local assembly + solve', 'Coarse assembly']),
        xcol=xcol,
        series_name=series_name,
        logx_base=2,
        logy_base=10,
        lgd_loc=6,
    )


def plot_fem(current, filename_base, series_name=None, xcol=None):
    xcol = xcol or 'cores'
    series_name = series_name or 'speedup'
    categories = ['apply', 'solve', 'constraints', 'assemble']
    ycols = ['fem.{}_avg_wall_speedup'.format(v) for v in categories] + ['ideal_speedup']
    labels = ['Overall', 'Solve', 'Constraints', 'Assembly', 'Ideal']
    plot_common(current, filename_base, ycols, labels, categories)


def plot_common(
    current,
    filename_base,
    ycols,
    labels,
    xcol,
    series_name,
    bar=None,
    logx_base=None,
    logy_base=None,
    color_map=None,
    bg_color=(1, 1, 1),
    lgd_loc=2,
):
    xlabels = {'cores': '\# Cores', 'grids.total_macro_cells': '\# Macro Cells'}
    fig = plt.figure()
    color_map = color_map or color_util.discrete_cmap(len(labels), bg_color=bg_color)
    subplot = current.plot(x=xcol, y=ycols, colormap=color_map, figsize=(6.3, 3.5))
    # [‘solid’ | ‘dashed’, ‘dashdot’, ‘dotted’ | (offset, on - off - dash - seq) | '-' | '--' | '-.' | ':' | 'None' | ' ' | '']

    for i, line in enumerate(subplot.lines):
        if i == len(subplot.lines) - 1:
            line.set_linestyle('dashed')
        line.set_marker(MARKERS[i])
    plt.ylabel(series_name.capitalize())
    plt.xlabel(xlabels[xcol])
    ax = subplot.figure.axes[0]
    if logx_base is not None:
        ax.set_xscale('log', basex=logx_base)
    if logy_base is not None:
        ax.set_yscale('log', basey=logy_base)
    lgd = plt.legend(ax.lines, labels, loc=lgd_loc)  # , bbox_to_anchor=(1.05, 1),  borderaxespad=0., loc=2)
    ax.set_facecolor(bg_color)
    lgd.get_frame().set_facecolor(bg_color)

    tt = current[xcol].array
    plt.xticks(ticks=tt, labels=[str(t) for t in tt])

    for fmt in FIGURE_OUTPUTS:
        plt.savefig(filename_base + '_{}.{}'.format(series_name, fmt), bbox_extra_artists=(lgd,), bbox_inches='tight')

    if bar is None:
        return
    cols, labels = bar
    fig = plt.figure()
    ax = current[cols].plot(kind='bar', stacked=True, colormap=color_map)
    patches, _ = ax.get_legend_handles_labels()

    lgd = ax.legend(patches, labels, bbox_to_anchor=(1.05, 1), borderaxespad=0.0, loc=2)

    plt.savefig(f'{filename_base}_{series_name}_pie.png', bbox_extra_artists=(lgd,), bbox_inches='tight')


def plot_error(
    data_frame, filename_base, error_cols, xcol, labels, baseline_name, logx_base=None, logy_base=None, color_map=None
):
    fig = plt.figure()

    # normed walltime
    normed = 'normed_walltime'
    w_time = data_frame['{}_avg_wall'.format(baseline_name)]
    count = len(w_time)
    values = [w_time[i] / w_time.max() for i in range(0, count)]
    # data_frame.insert(0, normed, pd.Series(values))

    color_map = color_map or color_util.discrete_cmap(len(labels))
    ax = data_frame.plot(x=xcol, y=error_cols, colormap=color_map)
    for i, line in enumerate(ax.lines):
        line.set_marker(MARKERS[i])
    plt.ylabel('Error')
    plt.xlabel('Cells')

    if logx_base is not None:
        ax.set_xscale('log', basex=logx_base)
    if logy_base is not None:
        ax.set_yscale('log', basey=logy_base)
    lgd = plt.legend(ax.lines, labels, loc=2)  # , bbox_to_anchor=(1.05, 1),  borderaxespad=0., loc=2)

    common = common_substring(error_cols)
    for fmt in FIGURE_OUTPUTS:
        plt.savefig('{}_{}.{}'.format(filename_base, common, fmt), bbox_extra_artists=(lgd,), bbox_inches='tight')



#!/usr/bin/env python3

import os
import sys
import itertools
import matplotlib as mpl

import pandas_common as pc


def plot_mlmc(current, filename_base):
    all = ['mlmc.all', 'fem.apply', 'msfem.Elliptic_MsFEM_Solver.apply']
    all_labels = ['Overall', 'CgFem', 'MsFem', 'Ideal']
    simple = ['mlmc.all']
    for pref, tpl in {'all': (all, all_labels), 'simple': (simple, ['Overall', 'Ideal'])}.items():
        cat, labels = tpl
        ycols = ['{}_avg_wall_speedup'.format(v) for v in cat] + ['ideal_speedup']
        pc.plot_common(current, '{}_{}'.format(pref, filename_base), ycols, labels)


common_string = pc.common_substring(sys.argv[1:])
merged = 'merged_{}.csv'.format(common_string)
baseline_name = 'mlmc.all'

header, current = pc.read_files(sys.argv[1:])
headerlist = header['profiler']
current = pc.sorted_f(current, True)
# full = pc.speedup(headerlist, current.copy(), baseline_name)
# full.transpose().to_csv(merged)
current = pc.speedup(headerlist, current, baseline_name)
# pprint(t_sections)
plot_mlmc(current, merged)

current.transpose().to_csv('filtered_' + merged)



#!/usr/bin/env python3

"""The setup script."""

from setuptools import setup, find_packages
import sys
import os

sys.path.append(os.path.dirname(__file__))
import dependencies
import versioneer

tests_require = dependencies.ci_requires
install_requires = dependencies.install_requires
setup_requires = dependencies.setup_requires()
install_suggests = dependencies.install_suggests

with open('README.rst') as readme_file:
    readme = readme_file.read()
with open('HISTORY.rst') as history_file:
    history = history_file.read()

setup(
    author="René Fritze",
    author_email='coding@fritze.me',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="MPI Aware section timings",
    entry_points={
        'console_scripts': [
            'pytimings=pytimings.cli:main',
        ],
    },
    tests_require=tests_require,
    install_requires=install_requires,
    extras_require=dependencies.extras(),
    license="BSD license",
    long_description=readme + '\n\n' + history + '\n\n',
    include_package_data=True,
    keywords='pytimings',
    name='pytimings',
    packages=find_packages(include=['pytimings', 'pytimings.*']),
    url='https://github.com/WWU-AMM/pytimings',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    zip_safe=False,
)



#!/usr/bin/env python3
import logging
import sys
import traceback

import pandas_common as pc

common_string = pc.common_substring(sys.argv[1:])
merged = 'merged_{}.csv'.format(common_string)
baseline_name = 'msfem.all'


def _get(func_name='scaleup'):
    header, current = pc.read_files(sys.argv[1:])
    headerlist = header['profiler']
    current = pc.sorted_f(current, True)
    func = getattr(pc, func_name)
    current = func(headerlist, current, baseline_name)
    current.transpose().to_csv('{}_{}'.format(merged, func_name))
    return current


current = _get('scaleup')
pc.plot_msfem(current, merged, series_name='scaleup', xcol='grids.total_macro_cells')

current = _get('speedup')
pc.plot_msfem(current, merged, series_name='speedup', xcol='cores')

try:
    pc.plot_error(
        data_frame=current,
        filename_base='{}_errors'.format(merged),
        error_cols=['msfem_exact_H1s', 'msfem_exact_L2', 'fem_exact_H1s', 'fem_exact_L2'],
        xcol='grids.micro_cells_per_macrocell_dim',
        labels=['MsFEM $H^1_s$', 'MsFEM $L^2$', 'CgFEM $H^1_s$', 'CgFEM $L^2$'],
        baseline_name=baseline_name,
        logy_base=10,
        logx_base=2,
    )
except KeyError as k:
    logging.error('No error data')
    logging.error(traceback.format_exc())
# current.transpose().to_excel(merged+'.xls')
# current.to_csv(merged)
